#%%
import pandas as pd
import numpy as np
pd.options.display.show_dimensions = False
pd.set_option('display.width', 1000)
pd.options.mode.copy_on_write = True
pd.set_option('display.max_rows', 15)
pd.set_option('display.max_columns', None)

import vertexai
from vertexai.generative_models import GenerativeModel
from NeuronsMemoryTestPipeline import constants
import NeuronsMemoryTestPipeline.MRT_free_recall_metric as MRT_free_recall_metric
import NeuronsMemoryTestPipeline.MRT_recognition_metric as MRT_recognition_metric

PROJECT = constants.PROJECT
LOCATION = constants.LOCATION

vertexai.init(project=PROJECT, location=LOCATION)
model = GenerativeModel(constants.MODEL_TYPE)

# ************** DOWNLOAD FILES FROM THE SCRIPT ************
############################################################    

#%%
def create_combined_score(freerecall, recognition, type=[]):
    wide_df = recognition.pivot_table(index=['group_id', 'mrt_stimulus'] + type, columns='mrt_association', values=['mrt_score_recognition'], aggfunc='first', observed=False)
    wide_df.columns = [f'{col[1]}' for col in wide_df.columns]
    wide_df.reset_index(drop=False, inplace=True)
    df_updated = freerecall.merge(wide_df, on = ['group_id', 'mrt_stimulus'] + type, how = 'outer')
    df_updated['Brand&Freerecall AVG'] = ((df_updated['Freerecall'] + df_updated['Which Brand?']) / 2)
    df_updated['Brand&Freerecall AVG'] = df_updated['Brand&Freerecall AVG'].clip(0, 100)
    #df_updated['Overall_Score'] = (alpha * (df_updated['Brand&Freerecall AVG']) + beta * df_updated['Remember Ad?']) / 2
    #df_updated['Overall_Score'] = df_updated['Overall_Score'].clip(0, 100)
    col1 = ['group_id', 'project_identifier', 'total_participants','mrt_stimulus']
    col2 = ['position_mean', 'brand_count', 'recall_%', 'Freerecall', 'Which Brand?', 'Brand&Freerecall AVG']
    df_updated = df_updated[col1 + type + col2].copy()
    return df_updated

def update_data(df, extra_df=None, type = None):
    return_df = pd.DataFrame()
    project = df['project_identifier'].dropna().unique()[0]
    for gr in df['group_id'].unique():
        check_df = df[df['group_id'] == gr].copy()
        uncheck_df = check_df[~check_df['project_identifier'].isna()]
        check_df = check_df[check_df['project_identifier'].isna()]
        check_df.project_identifier = project
        check_df['recall_%']=0
        check_df['position_mean']='na'
        check_df['brand_count']=0
        check_df['Freerecall']=0
        
        if type is not None and type in df.columns:
            check_df = check_df.merge(extra_df, how='left', on=['group_id', type])
            check_df['total_participants'] = check_df['count']
            check_df.drop('count', axis=1, inplace=True)
        else:
            check_df.drop('total_participants', axis=1, inplace=True)
            check_df = check_df.merge(extra_df, how='left', on='group_id')

        return_df = pd.concat([return_df, uncheck_df, check_df], axis=0, ignore_index=True)
        return_df.total_participants = return_df.total_participants.astype(int)
    return return_df

#%%
#PATHS and FILES FOR WINDOWS
#project_path = r'G:\Shared drives\HQ - Projects\Active Internal Projects\82025-0005_Brand_Recall_Carousel\\'
#path_output = project_path + r'Analysis_(DS)\Scores\Videos\\'
#cleaned_df_file = r"QA\Videos\Day2\G3\QA_Outputs\cleaned_df.csv"
#alias_map_file = "Target_Stimuli\Stimuli_Map\Videos\82025-0005_Brand_Recall_Carousel_Videos_map_G3.csv"

# for testing purposes
irina_path = r'/Users/irinakw/Library/CloudStorage/GoogleDrive-i.white@neuronsinc.com/Shared drives/HQ - Projects /'
project_path = irina_path + r'Active Internal Projects/82025-0005_Brand_Recall_Carousel/'
path_output = project_path + r'Analysis_(DS)/Scores/'
cleaned_df_file = f'QA/Videos/Day2/G3/QA_Outputs/cleaned_df.csv'
alias_map_file = "Target_Stimuli/Stimuli_Map/Videos/82025-0005_Brand_Recall_Carousel_Videos_map_G3.csv"
#demographics_df_file = r'2_Fieldwork/QA/Demographics_GoodIDs.xlsx'


#DOWNLOAD DEMOGRAPHICS, DATASET AND QA
df = pd.read_csv(project_path+cleaned_df_file, low_memory = True)
alias_data = pd.read_csv(project_path + alias_map_file, low_memory = True)

# Excluding bad IDs if necessary
# path_meta_RealEye = r"G:\Shared drives\HQ - Projects\Active External Projects\Standard Explore Projects\90696-0004_Teads_DE_TVAdTest_SamsungFold_AdEffectiveness\QA\Day7\TestGroup\ET\ET_QA_report.xlsx"
# df_meta_RE = pd.read_excel(path_meta_RealEye, sheet_name='Good IDs', header=0)
# df_meta_RE.rename(columns={"participant_id":'participant_id'}, inplace=True) # rename id column if necessary
# df_meta_RE['participant_id'] = df_meta_RE['participant_id'].astype(str)
# df = df[df['participant_id'].isin(df_meta_RE['participant_id'])]

# %% 
#CLEANED DATA
df.group_id = df.group_id.astype(int)
df.group_id = df.group_id.astype(str)
print("Group IDs in this study are: ", df.group_id.unique())
print('==='*30)
print('\nNumber of participants per group in the analyzed study:')
print(df.groupby('group_id')['participant_id'].nunique())
print('DATA is UPLOADED and READY for ANALYSIS!')
print("CHANGE THE GROUP_ID?")

# %% 
####################################################################################
# ************** POST SURVEY - FAMILIARITY, RELEVANCE, PREFERENCE **************
####################################################################################

try: 
    df_survey = df[df['trial_name'] == 'Survey_Question_Association_Data'].copy()

    if not df_survey.empty:
        df_survey['survey_question_association_text_english'] = df_survey['survey_question_association_text_english'].replace({
            'brand-familiarity': 'Brand Familiarity',
            'brand-preference': 'Brand Preference',
            'brand-relevance': 'Brand Relevance'
        })

        survey_pivot = df_survey.pivot_table(index = ['group_id', 'survey_question_association_text_english'],
                                            columns = 'trial_stimulus',
                                            values = 'survey_question_association_response_numeric',
                                            aggfunc = 'mean',
                                            fill_value = 0, observed = False)
        survey_pivot.index.name = None
        survey_pivot = survey_pivot.round(2)

        survey_exists = True
        print('\n ***** SURVEY DONE! *****')
    else:
        survey_exists = False
        print('\n ***** NO SURVEY DATA FOUND! *****')

except Exception as e: 
    survey_exists = False
    print(f'\n ***** AN ERROR OCCURRED IN SURVEY PROCESSING: {e} *****')
    pass

# if CHANGE_GROUP_ID not in df.group_id.unique():
#     print(f"Changed group_id to {CHANGE_GROUP_ID}")
#     df.group_id = CHANGE_GROUP_ID
#     df["group_id"] = pd.to_numeric(df["group_id"], errors="coerce").astype("Int64")
#     df["group_id"] = df["group_id"].astype(str)

#%%
# ************** MANUAL CORRECTION **************
#################################################
manual_correction = {
}
apply_manual_correction = False

# ************** PREPARE DATA **************
############################################
alias_data = MRT_free_recall_metric.alias_group_define(alias_data)
alias_data = alias_data.loc[alias_data['role']!='Distractor'].copy()
alias_data = alias_data[['group_id', 'brand_name', 'alias']].reset_index(drop=True).copy()
alias_data.columns = ['group_id', 'mrt_stimulus', 'mrt_stimulus_ad']
alias_data = alias_data[alias_data['group_id'].isin(df['group_id'].unique())].reset_index(drop=True).copy()
print('==='*30)
print('\nList of stimulus used in the study per analyzed Group:')
print(alias_data)

#%%
df = df[constants.MRT_COLUMNS]
df["Age_Group"] = pd.cut(x=df['Age'], bins=3, labels=["Younger","Mid","Older"])
df = df.loc[df['reason_to_end_the_behavioral_task_code']==13]
df.group_id = df.group_id.astype('str')
df.drop('reason_to_end_the_behavioral_task_code', axis = 1, inplace=True)
try: 
    df = df.loc[df['participant_quality'] == 1.0]
except: 
    pass
df['total_participants'] = df.groupby('group_id')['participant_id'].transform('nunique')
df = df.loc[df.module_name == 'MRT']

if apply_manual_correction:
    print('==='*30)
    print('\n***** Manual corrections to Free Recall applied *****\n')
    df['given_response_label_presented'] = df['given_response_label_presented'].replace(manual_correction)
else:
    print('==='*30)
    print('\nNo Manual Correction applied to the Free Recall entries.\n')

#%%
############################################
# ************** DEMOGRAPHICS **************
############################################
stat_age, stat_gender, stat_age_group = MRT_recognition_metric.produce_demographics(df)

stat_g = stat_gender.reset_index(drop=False)[['group_id','Gender','count']]
stat_g.columns = ['group_id','split','count']

stat_a = stat_age_group.reset_index(drop=False)[['group_id','Age_Group','count']]
stat_a.columns = ['group_id','split','count']

#%%
###########################################################
# ************** CALCULATE FREE RECALL SCORE **************
###########################################################

df_free_recall, corrected_pairs, not_match = MRT_free_recall_metric.process_recall(df,'Free Recall', model, alias_df = alias_data)
responses1=not_match[not_match['given_response_label_presented'].str.strip().str.split(' ').str.len() ==1]['given_response_label_presented'].reset_index(drop=True)
responses2=not_match[not_match['given_response_label_presented'].str.strip().str.split(' ').str.len() ==2]['given_response_label_presented'].reset_index(drop=True)
responses3plus=not_match[not_match['given_response_label_presented'].str.strip().str.split(' ').str.len() >2]['given_response_label_presented'].reset_index(drop=True)
corrected_pairs = corrected_pairs.drop_duplicates().dropna().sort_values(['method','given_response_label_presented']).reset_index(drop=True)
#%%
corrected_pairs
#%%
not_match
#%%
#Filter the responses we need to calculate
df_free_recall= df_free_recall[df_free_recall.corrected.isin(alias_data['mrt_stimulus'].unique())].reset_index(drop=True)
free_recall_scores = MRT_free_recall_metric.compute_scores(df_free_recall)
print('==='*30)
print('Free Recall scores are calculated!')
free_recall_scores


#%%
df_a = df_free_recall.merge(stat_a[['group_id', 'split', 'count']], how='left', left_on=['group_id','Age_Group'], right_on=['group_id','split'])
df_a['total_participants'] = df_a['count']
df_g = df_free_recall.merge(stat_g[['group_id', 'split', 'count']], how='left', left_on=['group_id', 'Gender'], right_on=['group_id','split'])
df_g['total_participants'] = df_g['count']
#%%
# DEMOGRAPHICS RESULTS FOR FREE RECALL
split_column_group = ['group_id', 'project_identifier', 'task_name', 'total_participants', 'corrected']
free_recall_scores_age = MRT_free_recall_metric.compute_scores(df_a, group = split_column_group + ['Age_Group'])
free_recall_scores_gender = MRT_free_recall_metric.compute_scores(df_g, group = split_column_group + ['Gender'])

# %%
############################################################
# **************  CALCULATE RECOGNITION RATES **************
############################################################

df_ad_recognition = df.loc[(df['task_name'] =='Ad Recognition') & (df['trial_name'] =='Presentation')].reset_index(drop=True)
df_ad_recognition['given_response_accuracy'] = df_ad_recognition['given_response_accuracy'].map({'Wrong': 'No', 'Correct': 'Yes'})

#%%
df_brand_recognition = df.loc[(df['task_name'] =='Ad Recognition') & (df['trial_name'] =='Recognition')].reset_index(drop=True)
df_brand_recognition['given_response_accuracy'] = df_brand_recognition['given_response_accuracy'].map({'Wrong': 'No', 'Correct': 'Yes'})

df_ad_recognition['mrt_association'] = 'Remember Ad?'
df_brand_recognition['mrt_association'] = 'Which Brand?'

mrt_column_rename = { 
    'expected_response' : 'mrt_stimulus',
    'given_response_accuracy': 'mrt_response'
    }
df_recognition = pd.concat([df_ad_recognition, df_brand_recognition], axis=0, ignore_index=True)
df_recognition.rename(columns = mrt_column_rename, inplace = True)
df_recognition_all = MRT_recognition_metric.recognition_scores(df_recognition).reset_index(drop=False)
df_recognition_all.rename(columns={'mrt_score': 'mrt_score_recognition'}, inplace=True)
df_recognition_gender = MRT_recognition_metric.split_score(df_recognition, ['Male', 'Female'], 'Gender').reset_index(drop=False)
df_recognition_gender.rename(columns={'mrt_score': 'mrt_score_recognition'}, inplace=True)
df_recognition_age = MRT_recognition_metric.split_score(df_recognition, ["Younger","Mid","Older"], 'Age_Group').reset_index(drop=False)
df_recognition_age.rename(columns={'mrt_score': 'mrt_score_recognition'}, inplace=True)



#%%
df_brand_recognition_dis = df_brand_recognition.drop_duplicates(subset=['expected_response', 'participant_id'])
df_brand_recognition_dis = df_brand_recognition.groupby(['group_id', 'expected_response', 'given_response_label_presented']).size().reset_index(name='count')
df_brand_recognition_dis['percentage'] = df_brand_recognition_dis.groupby(['group_id', 'expected_response'])['count'].transform(lambda x: (x / x.sum()) * 100)

#%% 
############################################################
# **************  MERGING ALL SCORES INTO ONE **************
############################################################
col = ['group_id', 'mrt_association', 'mrt_stimulus', 'mrt_score_recognition']

def filter_result(df, alias_data, col):
    merged_df = df.merge(alias_data, left_on=['group_id', 'mrt_stimulus'], right_on=['group_id', 'mrt_stimulus_ad'], how='left', suffixes=('', '_alias'))
    merged_df['mrt_stimulus'] = merged_df['mrt_stimulus_alias'].combine_first(merged_df['mrt_stimulus'])
    df_filtered = merged_df[col].copy()
    return df_filtered

df_recognition_all = filter_result(df_recognition_all, alias_data, col)
df_recognition_all = df_recognition_all.sort_values('group_id').reset_index(drop=True)
df_recognition_age = filter_result(df_recognition_age, alias_data, col = col+['Age_Group'])
df_recognition_gender = filter_result(df_recognition_gender, alias_data, col = col+['Gender'])

final_all = create_combined_score(free_recall_scores, df_recognition_all)
final_all_gender = create_combined_score(free_recall_scores_gender, df_recognition_gender, type = ['Gender'])
final_all_age = create_combined_score(free_recall_scores_age, df_recognition_age, type = ['Age_Group'])

#check for empty rows
age= stat_age_group[['group_id','Age_Group','count']].copy()
gender= stat_gender.reset_index(drop=False)[['group_id','Gender','count']].copy()
total_number = stat_gender.reset_index(drop=False)[['group_id','total_participants']].drop_duplicates().reset_index(drop=True).copy()
final_all = update_data(final_all, total_number)
final_all_age = update_data(final_all_age, age, 'Age_Group')
final_all_gender = update_data(final_all_gender, gender, 'Gender')


#%%
###################################################
# ************** AVERAGE POSITIONING **************
###################################################
print('==='*30)
df_entries = df[(df.task_name == 'Free Recall' )& (df.given_response_label_presented.notna())].copy()
df_free_recall[df_free_recall.participant_id == '3sM8K44fLRcUoPrenHHhMH'].reset_index(drop=True)
ratio_df = pd.DataFrame(index=df_free_recall['participant_id'].value_counts().index)
ratio_df['Targeted Entries'] = df_free_recall['participant_id'].value_counts()
ratio_df['All Entries'] = df_entries['participant_id'].value_counts()
ratio_df['Ratio'] = ratio_df['Targeted Entries'] / ratio_df['All Entries']
ratio_df = ratio_df.rename(columns={'index': 'participant_id'}).reset_index()
ratio_df = ratio_df.merge(df[['participant_id', 'group_id']].drop_duplicates(), on='participant_id', how='left')
ratio_df = ratio_df.groupby('group_id').agg({'Ratio': 'mean', 'Targeted Entries': 'mean', 'All Entries': 'mean'})

#%%
##########################################
# **************  REPORTING **************
##########################################
#path_output = ''  # For testing purposes

df_fr = pd.DataFrame({'Data':["Free Recall correction using fuzzy and AI"]})
with pd.ExcelWriter(path_output + f'MRT_Scores_Videos_G3.xlsx') as writer:
    stat_age.to_excel(writer,sheet_name = 'Demographics', index = True, startcol = 0)
    stat_age_group.to_excel(writer,sheet_name = 'Demographics', index = False, startcol = 6)
    stat_gender.to_excel(writer,sheet_name = 'Demographics', index = True, startcol = 14)
    final_all.set_index(['group_id', 'mrt_stimulus']).to_excel(writer, sheet_name = "MRT Scores Final", startrow = 0, startcol = 0, index = True) 
    ratio_df.to_excel(writer,sheet_name = 'MRT Scores Final', index = True, startrow = 0, startcol = 15)
    final_all_age.set_index(['group_id', 'mrt_stimulus','Age_Group']).to_excel(writer, sheet_name = "MRT Scores Final Age Split", startrow = 0, startcol = 0, index = True)
    final_all_gender.set_index(['group_id', 'mrt_stimulus','Gender']).to_excel(writer, sheet_name = "MRT Scores Final Gender Split", startrow = 0, startcol = 0, index = True)
    if survey_exists:
        survey_pivot.to_excel(writer, sheet_name = "Post-Survey", index = True)

with pd.ExcelWriter(path_output + f'MRT_ExtraSheets_Videos_G3.xlsx') as writer:
    alias_data.to_excel(writer,sheet_name = 'Stimulus', index=False)
    corrected_pairs.set_index(['group_id', 'method']).to_excel(writer, sheet_name = "Free Text Correction", startrow = 0, startcol = 0, index = True)  
    responses1.to_excel(writer, sheet_name = "FreeRecall Responses", index = False, startcol = 0)
    responses2.to_excel(writer, sheet_name = "FreeRecall Responses", index = False, startcol = 3) 
    responses3plus.to_excel(writer, sheet_name = "FreeRecall Responses", index = False, startcol = 6)
    df_brand_recognition_dis.to_excel(writer, sheet_name = "Brand Recognition Responses", index = False, startcol = 0)


print('***** MRT REPORT IS DONE! *****')
# %%



# ==== Fast, grounded-first brand cleaner (final) ====
import re, unicodedata
from collections import defaultdict
import pandas as pd
from rapidfuzz import process, fuzz
from rapidfuzz.distance import Levenshtein
from tqdm.auto import tqdm
from NeuronsMemoryTestPipeline.constants import SPELLCHECK_PROMPT_TEMPLATE

# ------------------- helpers -------------------
GENERIC_NOT_FOUND = {
    "idk","i don't know","dont know","can't remember","cant remember",
    "no idea","nothing","na","n/a","none","other","others"
}

def normalize(s: str) -> str:
    if not isinstance(s, str): return ""
    s = unicodedata.normalize("NFKC", s)
    s = s.strip()
    s = re.sub(r"\s+", " ", s)     
    s = s.lower()
    s = re.sub(r"[^\w\s&+]", "", s)  
    return s

def nospace(s: str) -> str: return re.sub(r"\s+", "", s)
def and_equiv(s: str) -> str: return re.sub(r"\band\b", "&", s)
def compress_runs(s: str) -> str: return re.sub(r"(.)\1+", r"\1", s)
def tokens(s: str):
    return [t for t in re.split(r"\s+", s.strip()) if t]

def looks_like_acronym_input(s: str) -> bool:
    s = s.strip()
    if len(s) < 2 or len(s) > 6:
        return False
    no_space = " " not in s
    has_dots = "." in s
    mostly_upper = sum(c.isupper() for c in s) >= max(1, int(0.6*len([c for c in s if c.isalpha()])))
    return (no_space and (has_dots or mostly_upper)) or has_dots

def abbr_of(phrase: str) -> str:
    return "".join(w[0].upper() for w in re.split(r"\s+", (phrase or "").strip()) if w)

def bigrams(s: str):
    return {s[i:i+2] for i in range(len(s)-1)} if len(s) > 1 else set()

# ------------------- indices -------------------
def build_indices(brands_presented):
    brands_presented = list(dict.fromkeys(brands_presented))  
    norm_to_orig = {normalize(b): b for b in brands_presented}
    norm_names = list(norm_to_orig.keys())

    abbr_map = defaultdict(list)
    for b in brands_presented:
        abbr_map[abbr_of(b)].append(b)

    # exact-equivalence maps
    ns_map  = {nospace(k): v for k, v in norm_to_orig.items()}
    and_map = {nospace(and_equiv(k)): v for k, v in norm_to_orig.items()}
    cr_map  = {compress_runs(k): v for k, v in norm_to_orig.items()}

    cr_ns_map = {nospace(compress_runs(k)): v for k, v in norm_to_orig.items()}
    ns_cr_map = {compress_runs(nospace(k)): v for k, v in norm_to_orig.items()}
    
    return (brands_presented, norm_to_orig, norm_names, abbr_map,
            ns_map, and_map, cr_map, cr_ns_map, ns_cr_map)

# ------------------- grounded match -------------------
def grounded_match(user_text: str,
                   norm_to_orig, norm_names, abbr_map,
                   ns_map, and_map, cr_map, cr_ns_map, ns_cr_map):
    raw = user_text or ""
    raw_norm = normalize(raw)

    # obvious not-found
    if raw_norm in GENERIC_NOT_FOUND or raw_norm == "":
        return "", "not_match"

    # exact normalized
    if raw_norm in norm_to_orig:
        return norm_to_orig[raw_norm], "exact"
    #token match
    toks = tokens(raw_norm)
    if toks:
        tok_set = set(toks)
        for bn, orig in norm_to_orig.items():
            if bn in tok_set:
                return orig, "brand_token"
            
    # exact-equivalence: nospace / and<->& / compressed runs
    k = nospace(raw_norm)
    if k in ns_map: return ns_map[k], "exact_nospace"

    k2 = nospace(and_equiv(raw_norm))
    if k2 in and_map: return and_map[k2], "exact_and_equiv"

    k3 = compress_runs(raw_norm)
    if k3 in cr_map: return cr_map[k3], "exact_compress"
            
    # NEW: combined exact maps (handles "go dady" -> "godaddy" via compress + nospace)
    k4 = nospace(compress_runs(raw_norm))
    if k4 in cr_ns_map: return cr_ns_map[k4], "exact_compress_nospace"
    k5 = compress_runs(nospace(raw_norm))
    if k5 in ns_cr_map: return ns_cr_map[k5], "exact_nospace_compress"

    # abbreviation (TIGHTENED without removing prior logic)
    ab = abbr_of(raw)
    # NEW: require the input to look like an acronym; ignore 1-letter abbreviations
    if looks_like_acronym_input(raw) and len(ab) >= 2:
        if ab in abbr_map and len(abbr_map[ab]) == 1:
            return abbr_map[ab][0], "abbreviation"
        elif ab in abbr_map and len(abbr_map[ab]) > 1:
            # NEW: avoid risky tie-breaks — defer to AI instead of forcing a brand
            return "", "maybe"

    # fuzzy (strict, short-string aware)
    choice, score, _ = process.extractOne(raw_norm, norm_names, scorer=fuzz.ratio)
    if not choice:
        return "", "maybe"

    cand_norm = choice
    cand_orig = norm_to_orig[cand_norm]
    Lx, Ly = len(raw_norm), len(cand_norm)
    len_delta = abs(Lx - Ly)
    edit_dist = Levenshtein.distance(raw_norm, cand_norm)
    bx = bigrams(raw_norm); by = bigrams(cand_norm)
    bigram_overlap = len(bx & by)
    first_letter_ok = (Lx == 0 or cand_norm[:1] == raw_norm[:1])

    if Lx <= 4:
        pass_threshold = (score >= 97) and (edit_dist <= 1) and first_letter_ok and (len_delta <= 1) and (bigram_overlap >= 1)
    elif Lx <= 6:
        pass_threshold = (score >= 94) and (edit_dist <= 2) and first_letter_ok and (len_delta <= 2) and (bigram_overlap >= 1)
    else:
        pass_threshold = (score >= 90) and (edit_dist <= 3) and (len_delta <= 3)

    if pass_threshold:
        return cand_orig, "fuzzy"
    return "", "maybe"

# ------------------- junk filters (numbers/gibberish) -------------------
VOWELS = set("aeiou")

def is_numeric_like(s: str) -> bool:
    s = (s or "").strip()
    if not s: return True
    d = sum(c.isdigit() for c in s)
    return s.isdigit() or (len(s) >= 4 and d / len(s) >= 0.5)

def looks_like_gibberish(s: str) -> bool:
    t = re.sub(r"[^a-z]", "", normalize(s))  # letters only
    if len(t) == 0: return True
    if len(t) >= 6 and not (set(t) & VOWELS): return True
    if len(t) >= 7 and (sum(ch in VOWELS for ch in t) / len(t) <= 0.1): return True
    if re.search(r"(.)\1{3,}", t): return True
    raw = normalize(s)
    non_alnum = sum(not c.isalnum() and not c.isspace() for c in raw)
    return len(raw) >= 6 and non_alnum / len(raw) > 0.3

def is_candidate_for_ai(s: str) -> bool:
    s = (s or "").strip()
    return bool(s) and s.lower() not in GENERIC_NOT_FOUND and not is_numeric_like(s) and not looks_like_gibberish(s)

# ------------------- ADDED: AI distance prefilter -------------------
# configurable gates (kept separate; nothing removed)
AI_MIN_RATIO     = 75      # min fuzz.ratio to nearest brand
AI_MAX_EDIT_FRAC = 0.25    # max Levenshtein distance as fraction of input length
AI_MIN_BIGRAMS   = 1       # require >=1 bigram overlap for len>=4

def _bigrams(s: str):
    return {s[i:i+2] for i in range(len(s)-1)} if len(s) > 1 else set()

def close_enough_to_brand(s: str, norm_names) -> bool:
    """Additional distance-based gate to avoid very-far pairs (e.g., 'Volvo' vs 'Citroen')."""
    sx = normalize(s)
    if not sx: return False
    hit = process.extractOne(sx, norm_names, scorer=fuzz.ratio)
    if not hit: return False
    cand_norm, ratio, _ = hit
    edit = Levenshtein.distance(sx, cand_norm)
    L = len(sx)
    max_edit = max(2, int(round(AI_MAX_EDIT_FRAC * L)))
    bg_ok = (L < 4) or (len(_bigrams(sx) & _bigrams(cand_norm)) >= AI_MIN_BIGRAMS)
    first_letter_ok = (L == 0) or (cand_norm[:1] == sx[:1])
    return (ratio >= AI_MIN_RATIO) and (edit <= max_edit) and bg_ok and first_letter_ok

# ------------------- AI fallback -------------------
def ai_spellcheck_batch(candidates_unique, brands_presented, norm_names, model, prompt_template, max_words=4):
    clean_candidates = [c for c in candidates_unique if is_candidate_for_ai(c)]
    near_candidates = [c for c in clean_candidates if close_enough_to_brand(c, norm_names)]
    print(f"[AI] candidates (post-filter): {len(near_candidates)} / {len(candidates_unique)} "
          f"(junk-kept: {len(clean_candidates)})")
    if (model is None) or (len(near_candidates) == 0):
        print(f"AI skipped. Model={model is not None}, candidates kept={len(near_candidates)}")
        return pd.DataFrame(columns=["given_response_label_presented","corrected","method"])

    rows = []
    for item in tqdm(near_candidates, desc="AI spellcheck", leave=False):
        prompt = prompt_template.format(
            potential_responses=brands_presented,
            brand_to_check=item
        )
        try:
            res = model.generate_content(prompt)
            txt = (res.candidates[0].text or "").strip()
        except Exception:
            txt = "Not Found"

        if len(txt.split()) > max_words:
            txt = "Not Found"

        if txt != "Not Found":
            rows.append({
                "given_response_label_presented": item,
                "corrected": txt,
                "method": "ai"
            })
    return pd.DataFrame(rows)

# ------------------- main CALL -------------------
def clean_free_recall_fast(df: pd.DataFrame, brands_presented, model=None, use_ai=True):
    (brands_presented, norm_to_orig, norm_names, abbr_map,
        ns_map, and_map, cr_map, cr_ns_map, ns_cr_map) = build_indices(brands_presented)

    out = df.copy()
    if "given_response_label_presented" not in out.columns:
        raise ValueError("Input df must have 'given_response_label_presented'")

    out["given_response_label_presented"] = out["given_response_label_presented"].astype(str).str.strip()
    out.loc[out["given_response_label_presented"].eq(""), "given_response_label_presented"] = pd.NA

    out["method"] = ""
    out["corrected"] = ""

    results = out["given_response_label_presented"].apply(
        lambda x: grounded_match(x, norm_to_orig, norm_names, abbr_map, ns_map, and_map, cr_map, cr_ns_map, ns_cr_map)
        if pd.notna(x) else ("","not_match")
    )
    out[["corrected","method"]] = pd.DataFrame(results.tolist(), index=out.index)

    mask_others = out["given_response_label_presented"].astype(str).str.lower().isin({"other","others"})
    out.loc[mask_others, "method"] = "not_match"

    if use_ai:
        if model is None:
            raise ValueError("use_ai=True but model is None")
        maybes = (out.loc[out["method"] == "maybe", "given_response_label_presented"]
                    .dropna().map(str).map(str.strip))
        maybes = [m for m in maybes if m]
        ai_df = ai_spellcheck_batch(maybes, brands_presented, norm_names, model, SPELLCHECK_PROMPT_TEMPLATE)

        if not ai_df.empty:
            out = out.merge(ai_df, on="given_response_label_presented", how="left", suffixes=("","_ai"))
            fill = out["corrected_ai"].notna()
            out.loc[fill, "corrected"] = out.loc[fill, "corrected_ai"]
            out.loc[fill, "method"]    = "ai"
            out = out.drop(columns=["corrected_ai","method_ai"], errors="ignore")

        out.loc[out["method"] == "maybe", "method"] = "not_match"

    return out


def calculate_corrected_brands_fast(df: pd.DataFrame, brands_presented):
    valid = set(brands_presented)
    count_df = df[df["corrected"].isin(valid)].drop_duplicates(["participant_id","corrected","method"])
    return (
        count_df.groupby(["group_id","method"])["corrected"]
        .value_counts()
        .rename("count")
        .reset_index()
    )
